@@ -116,11 +116,13 @@ import os
 import time
 import os.path
 from typing import (
-    Dict, List, Set, Tuple, Iterable, Union, Optional, Mapping, NamedTuple, Callable
+    Dict, List, Set, Tuple, Iterable, Union, Optional, Mapping, NamedTuple, Callable,
+    Sequence,
 )
 
 from mypy.build import (
     BuildManager, State, BuildSource, BuildResult, Graph, load_graph, module_not_found,
+    process_fresh_modules,
     PRI_INDIRECT, DEBUG_FINE_GRAINED,
 )
 from mypy.checker import DeferredNode
@@ -154,12 +156,12 @@ class FineGrainedBuildManager:
             result: Result from the initialized build.
                     The manager and graph will be taken over by this class.
             manager: State of the build (mutated by this class)
-            graph: Additional state of the build (only read to initialize state)
+            graph: Additional state of the build (mutated by this class)
         """
         manager = result.manager
         self.manager = manager
         self.graph = result.graph
-        self.previous_modules = get_module_to_path_map(manager)
+        self.previous_modules = get_module_to_path_map(self.graph)
         self.deps = get_all_dependencies(manager, self.graph)
         self.previous_targets_with_errors = manager.errors.targets()
         self.previous_messages = result.errors[:]
@@ -188,7 +190,7 @@ class FineGrainedBuildManager:
         those parts of other modules that are affected by the changes. Retain
         the existing ASTs and symbol tables of unaffected modules.
 
-        Create new graph with new State objects, but reuse original BuildManager.
+        Reuses original BuildManager and Graph.
 
         Args:
             changed_modules: Modules changed since the previous update/build; each is
@@ -324,6 +326,10 @@ class FineGrainedBuildManager:
         previous_modules = self.previous_modules
         graph = self.graph
 
+        # If this is an already existing module, make sure that we have
+        # its tree loaded so that we can snapshot it for comparison.
+        ensure_trees_loaded(manager, graph, [module])
+
         # Record symbol table snaphot of old version the changed module.
         old_snapshots = {}  # type: Dict[str, Dict[str, SnapshotItem]]
         if module in manager.modules:
@@ -336,7 +342,7 @@ class FineGrainedBuildManager:
         if isinstance(result, BlockedUpdate):
             # Blocking error -- just give up
             module, path, remaining, errors = result
-            self.previous_modules = get_module_to_path_map(manager)
+            self.previous_modules = get_module_to_path_map(graph)
             return remaining, (module, path), errors
         assert isinstance(result, NormalUpdate)  # Work around #4124
         module, path, remaining, tree = result
@@ -348,7 +354,7 @@ class FineGrainedBuildManager:
                         if not trigger.endswith('__>')]
             self.manager.log_fine_grained('triggered: %r' % sorted(filtered))
         self.triggered.extend(triggered | self.previous_targets_with_errors)
-        collect_dependencies({module: tree}, self.deps, graph)
+        collect_dependencies([module], self.deps, graph)
         remaining += propagate_changes_using_dependencies(
             manager, graph, self.deps, triggered,
             {module},
@@ -356,16 +362,54 @@ class FineGrainedBuildManager:
 
         # Preserve state needed for the next update.
         self.previous_targets_with_errors.update(manager.errors.targets())
-        self.previous_modules = get_module_to_path_map(manager)
+        self.previous_modules = get_module_to_path_map(graph)
 
         return remaining, (module, path), None
 
 
+def find_unloaded_deps(manager: BuildManager, graph: Dict[str, State],
+                       initial: Sequence[str]) -> List[str]:
+    """Find all the deps of the nodes in initial that haven't had their tree loaded.
+
+    The key invariant here is that if a module is loaded, so are all
+    of their dependencies. This means that when we encounter a loaded
+    module, we don't need to explore its dependencies.  (This
+    invariant is slightly violated when dependencies are added, which
+    can be handled by calling find_unloaded_deps directly on the new
+    dependencies.)
+    """
+    worklist = list(initial)
+    seen = set()  # type: Set[str]
+    unloaded = []
+    while worklist:
+        node = worklist.pop()
+        if node in seen or node not in graph:
+            continue
+        seen.add(node)
+        if node not in manager.modules:
+            ancestors = graph[node].ancestors or []
+            worklist.extend(graph[node].dependencies + ancestors)
+            unloaded.append(node)
+
+    return unloaded
+
+
+def ensure_trees_loaded(manager: BuildManager, graph: Dict[str, State],
+                        initial: Sequence[str]) -> None:
+    """Ensure that the modules in initial and their deps have loaded trees."""
+    to_process = find_unloaded_deps(manager, graph, initial)
+    if to_process:
+        if is_verbose(manager):
+            manager.log_fine_grained("Calling process_fresh_modules on set of size {} ({})".format(
+                len(to_process), sorted(to_process)))
+        process_fresh_modules(graph, to_process, manager)
+
+
 def get_all_dependencies(manager: BuildManager, graph: Dict[str, State]) -> Dict[str, Set[str]]:
     """Return the fine-grained dependency map for an entire build."""
     # Deps for each module were computed during build() or loaded from the cache.
     deps = {}  # type: Dict[str, Set[str]]
-    collect_dependencies(manager.modules, deps, graph)
+    collect_dependencies(graph, deps, graph)
     return deps
 
 
@@ -413,40 +457,56 @@ def update_module_isolated(module: str,
 
     Returns a named tuple describing the result (see above for details).
     """
-    if module not in manager.modules:
+    if module not in graph:
         manager.log_fine_grained('new module %r' % module)
 
     if not manager.fscache.isfile(path) or force_removed:
         delete_module(module, graph, manager)
         return NormalUpdate(module, path, [], None)
 
-    old_modules = dict(manager.modules)
     sources = get_sources(manager.fscache, previous_modules, [(module, path)])
 
     if module in manager.missing_modules:
         manager.missing_modules.remove(module)
 
+    orig_module = module
+    orig_state = graph.get(module)
+    orig_tree = manager.modules.get(module)
+
+    def restore(ids: List[str]) -> None:
+        # For each of the modules in ids, restore that id's old
+        # manager.modules and graphs entries. (Except for the original
+        # module, this means deleting them.)
+        for id in ids:
+            if id == orig_module and orig_tree:
+                manager.modules[id] = orig_tree
+            elif id in manager.modules:
+                del manager.modules[id]
+            if id == orig_module and orig_state:
+                graph[id] = orig_state
+            elif id in graph:
+                del graph[id]
+
+    new_modules = []  # type: List[State]
     try:
         if module in graph:
             del graph[module]
-        load_graph(sources, manager, graph)
+        load_graph(sources, manager, graph, new_modules)
     except CompileError as err:
         # Parse error somewhere in the program -- a blocker
         assert err.module_with_blocker
-        if err.module_with_blocker != module:
-            # Blocker is in a fresh module. Delete the state of the original target module
-            # since it will be stale.
-            #
-            # TODO: It would be more efficient to store the original target module
-            path = manager.modules[module].path
-            del manager.modules[module]
-            remaining_modules = [(module, path)]
-        else:
-            remaining_modules = []
-        return BlockedUpdate(err.module_with_blocker, path, remaining_modules, err.messages)
+        restore([module] + [st.id for st in new_modules])
+        return BlockedUpdate(err.module_with_blocker, path, [], err.messages)
+
+    # Reparsing the file may have brought in dependencies that we
+    # didn't have before. Make sure that they are loaded to restore
+    # the invariant that a module having a loaded tree implies that
+    # its dependencies do as well.
+    ensure_trees_loaded(manager, graph, graph[module].dependencies)
 
     # Find any other modules brought in by imports.
-    changed_modules = get_all_changed_modules(module, path, previous_modules, graph)
+    changed_modules = [(st.id, st.xpath) for st in new_modules]
+
     # If there are multiple modules to process, only process one of them and return
     # the remaining ones to the caller.
     if len(changed_modules) > 1:
@@ -455,12 +515,7 @@ def update_module_isolated(module: str,
         changed_modules.remove((module, path))
         remaining_modules = changed_modules
         # The remaining modules haven't been processed yet so drop them.
-        for id, _ in remaining_modules:
-            if id in old_modules:
-                manager.modules[id] = old_modules[id]
-            else:
-                del manager.modules[id]
-            del graph[id]
+        restore([id for id, _ in remaining_modules])
         manager.log_fine_grained('--> %r (newly imported)' % module)
     else:
         remaining_modules = []
@@ -474,17 +529,15 @@ def update_module_isolated(module: str,
         state.semantic_analysis()
     except CompileError as err:
         # There was a blocking error, so module AST is incomplete. Restore old modules.
-        manager.modules.clear()
-        manager.modules.update(old_modules)
-        del graph[module]
+        restore([module])
         return BlockedUpdate(module, path, remaining_modules, err.messages)
     state.semantic_analysis_pass_three()
     state.semantic_analysis_apply_patches()
 
     # Merge old and new ASTs.
     assert state.tree is not None, "file must be at least parsed"
-    new_modules = {module: state.tree}  # type: Dict[str, Optional[MypyFile]]
-    replace_modules_with_new_variants(manager, graph, old_modules, new_modules)
+    new_modules_dict = {module: state.tree}  # type: Dict[str, Optional[MypyFile]]
+    replace_modules_with_new_variants(manager, graph, {orig_module: orig_tree}, new_modules_dict)
 
     # Perform type checking.
     state.type_checker().reset()
@@ -555,9 +608,9 @@ def dedupe_modules(modules: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
     return result
 
 
-def get_module_to_path_map(manager: BuildManager) -> Dict[str, str]:
-    return {module: node.path
-            for module, node in manager.modules.items()}
+def get_module_to_path_map(graph: Graph) -> Dict[str, str]:
+    return {module: node.xpath
+            for module, node in graph.items()}
 
 
 def get_sources(fscache: FileSystemCache,
@@ -570,25 +623,11 @@ def get_sources(fscache: FileSystemCache,
     return sources
 
 
-def get_all_changed_modules(root_module: str,
-                            root_path: str,
-                            old_modules: Dict[str, str],
-                            new_graph: Dict[str, State]) -> List[Tuple[str, str]]:
-    changed_set = {root_module}
-    changed_modules = [(root_module, root_path)]
-    for st in new_graph.values():
-        if st.id not in old_modules and st.id not in changed_set:
-            assert st.path
-            changed_set.add(st.id)
-            changed_modules.append((st.id, st.path))
-    return changed_modules
-
-
-def collect_dependencies(new_modules: Mapping[str, Optional[MypyFile]],
+def collect_dependencies(new_modules: Iterable[str],
                          deps: Dict[str, Set[str]],
                          graph: Dict[str, State]) -> None:
-    for id, node in new_modules.items():
-        if node is None:
+    for id in new_modules:
+        if id not in graph:
             continue
         for trigger, targets in graph[id].fine_grained_deps.items():
             deps.setdefault(trigger, set()).update(targets)
@@ -639,7 +678,7 @@ def calculate_active_triggers(manager: BuildManager,
 def replace_modules_with_new_variants(
         manager: BuildManager,
         graph: Dict[str, State],
-        old_modules: Dict[str, MypyFile],
+        old_modules: Dict[str, Optional[MypyFile]],
         new_modules: Dict[str, Optional[MypyFile]]) -> None:
     """Replace modules with newly builds versions.
 
@@ -651,10 +690,10 @@ def replace_modules_with_new_variants(
     propagate_changes_using_dependencies).
     """
     for id in new_modules:
+        preserved_module = old_modules.get(id)
         new_module = new_modules[id]
-        if id in old_modules and new_module is not None:
-            preserved_module = old_modules[id]
-            merge_asts(preserved_module, old_modules[id].names,
+        if preserved_module and new_module is not None:
+            merge_asts(preserved_module, preserved_module.names,
                        new_module, new_module.names)
             manager.modules[id] = preserved_module
             graph[id].tree = preserved_module
@@ -673,7 +712,7 @@ def propagate_changes_using_dependencies(
     a target that needs to be reprocessed but that has not been parsed yet."""
 
     num_iter = 0
-    remaining_modules = []
+    remaining_modules = []  # type: List[Tuple[str, str]]
 
     # Propagate changes until nothing visible has changed during the last
     # iteration.
@@ -682,11 +721,14 @@ def propagate_changes_using_dependencies(
         if num_iter > MAX_ITER:
             raise RuntimeError('Max number of iterations (%d) reached (endless loop?)' % MAX_ITER)
 
-        todo = find_targets_recursive(manager, triggered, deps, up_to_date_modules)
+        todo, unloaded = find_targets_recursive(manager, graph,
+                                                triggered, deps, up_to_date_modules)
+        # TODO: we sort to make it deterministic, but this is *incredibly* ad hoc
+        remaining_modules.extend((id, graph[id].xpath) for id in sorted(unloaded))
         # Also process targets that used to have errors, as otherwise some
         # errors might be lost.
         for target in targets_with_errors:
-            id = module_prefix(manager.modules, target)
+            id = module_prefix(graph, target)
             if id is not None and id not in up_to_date_modules:
                 if id not in todo:
                     todo[id] = set()
@@ -696,13 +738,7 @@ def propagate_changes_using_dependencies(
         # TODO: Preserve order (set is not optimal)
         for id, nodes in sorted(todo.items(), key=lambda x: x[0]):
             assert id not in up_to_date_modules
-            if manager.modules[id].is_cache_skeleton:
-                # We have only loaded the cache for this file, not the actual file,
-                # so we can't access the nodes to reprocess.
-                # Add it to the queue of files that need to be processed fully.
-                remaining_modules.append((id, manager.modules[id].path))
-            else:
-                triggered |= reprocess_nodes(manager, graph, id, nodes, deps)
+            triggered |= reprocess_nodes(manager, graph, id, nodes, deps)
         # Changes elsewhere may require us to reprocess modules that were
         # previously considered up to date. For example, there may be a
         # dependency loop that loops back to an originally processed module.
@@ -716,16 +752,21 @@ def propagate_changes_using_dependencies(
 
 def find_targets_recursive(
         manager: BuildManager,
+        graph: Graph,
         triggers: Set[str],
         deps: Dict[str, Set[str]],
-        up_to_date_modules: Set[str]) -> Dict[str, Set[DeferredNode]]:
+        up_to_date_modules: Set[str]) -> Tuple[Dict[str, Set[DeferredNode]],
+                                               Set[str]]:
     """Find names of all targets that need to reprocessed, given some triggers.
 
-    Returns: Dictionary from module id to a set of stale targets.
+    Returns: A tuple containing a:
+     * Dictionary from module id to a set of stale targets.
+     * A set of module ids for unparsed modules with stale targets.
     """
     result = {}  # type: Dict[str, Set[DeferredNode]]
     worklist = triggers
     processed = set()  # type: Set[str]
+    unloaded_files = set()  # type: Set[str]
 
     # Find AST nodes corresponding to each target.
     #
@@ -738,20 +779,28 @@ def find_targets_recursive(
             if target.startswith('<'):
                 worklist |= deps.get(target, set()) - processed
             else:
-                module_id = module_prefix(manager.modules, target)
+                module_id = module_prefix(graph, target)
                 if module_id is None:
                     # Deleted module.
                     continue
                 if module_id in up_to_date_modules:
                     # Already processed.
                     continue
+                if (module_id not in manager.modules
+                        or manager.modules[module_id].is_cache_skeleton):
+                    # We haven't actually parsed and checked the module, so we don't have
+                    # access to the actual nodes.
+                    # Add it to the queue of files that need to be processed fully.
+                    unloaded_files.add(module_id)
+                    continue
+
                 if module_id not in result:
                     result[module_id] = set()
                 manager.log_fine_grained('process: %s' % target)
                 deferred = lookup_target(manager, target)
                 result[module_id].update(deferred)
 
-    return result
+    return (result, unloaded_files)
 
 
 def reprocess_nodes(manager: BuildManager,
